"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from biweeklybudget.plaid_updater import PlaidUpdateResult, PlaidUpdater
from biweeklybudget.models.account import Account
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.models.plaid_items import PlaidItem
from biweeklybudget.models.plaid_accounts import PlaidAccount
from plaid.api.plaid_api import PlaidApi
from datetime import datetime, date
from decimal import Decimal
from pytz import UTC

from unittest.mock import Mock, patch, call, MagicMock, DEFAULT
import pytest

pbm = 'biweeklybudget.plaid_updater'
pb = f'{pbm}.PlaidUpdater'


class TestPlaidUpdateResult:

    def test_happy_path(self):
        item = Mock(spec_set=PlaidItem)
        type(item).item_id = 4
        r = PlaidUpdateResult(
            item, True, 1, 2, 'foo', [123]
        )
        assert r.item == item
        assert r.success is True
        assert r.updated == 1
        assert r.added == 2
        assert r.exc == 'foo'
        assert r.stmt_ids == [123]
        assert r.as_dict == {
            'item_id': 4,
            'success': True,
            'exception': 'foo',
            'statement_ids': [123],
            'added': 2,
            'updated': 1
        }


class TestInit:

    def test_happy_path(self):
        mock_plaid = Mock()
        with patch(f'{pbm}.plaid_client') as m_pc:
            m_pc.return_value = mock_plaid
            cls = PlaidUpdater()
        assert cls.client == mock_plaid
        assert m_pc.mock_calls == [call()]


class TestAvailableItems:

    def test_happy_path(self):
        items = [
            Mock(id=1),
            Mock(id=2)
        ]
        mock_name = Mock()
        mock_item = Mock(institution_name=mock_name)
        with patch(f'{pbm}.db_session') as mock_db:
            with patch(f'{pbm}.PlaidItem', mock_item):
                mock_db.query.return_value.order_by.return_value.all\
                    .return_value = items
                res = PlaidUpdater.available_items()
        assert res == items
        assert mock_db.mock_calls == [
            call.query(mock_item),
            call.query().order_by(mock_name),
            call.query().order_by().all()
        ]


class PlaidUpdaterTester:

    def setup_method(self):
        self.mock_client = MagicMock()
        with patch(f'{pbm}.plaid_client') as m_pc:
            m_pc.return_value = self.mock_client
            self.cls = PlaidUpdater()


class TestUpdate(PlaidUpdaterTester):

    def test_accounts_none(self):
        item1 = Mock()
        item2 = Mock()
        item3 = Mock()
        items = [item1, item2, item3]
        res1 = Mock()
        res2 = Mock()
        res3 = Mock()
        results = [res1, res2, res3]
        with patch(f'{pb}.available_items') as m_avail:
            m_avail.return_value = items
            with patch(f'{pb}._do_item') as m_do_item:
                m_do_item.side_effect = results
                res = self.cls.update()
        assert res == results
        assert m_avail.mock_calls == [call()]
        assert m_do_item.mock_calls == [
            call(item1, days=30),
            call(item2, days=30),
            call(item3, days=30)
        ]

    def test_accounts_specified(self):
        item1 = Mock()
        item2 = Mock()
        item3 = Mock()
        items = [item1, item2, item3]
        res1 = Mock()
        res3 = Mock()
        results = [res1, res3]
        with patch(f'{pb}.available_items') as m_avail:
            m_avail.return_value = items
            with patch(f'{pb}._do_item') as m_do_item:
                m_do_item.side_effect = results
                res = self.cls.update(items=[item1, item3], days=10)
        assert res == results
        assert m_avail.mock_calls == []
        assert m_do_item.mock_calls == [
            call(item1, days=10),
            call(item3, days=10)
        ]


class TestDoItem(PlaidUpdaterTester):

    def test_happy_path(self):
        acctA = Mock(spec_set=Account)
        acctB = Mock(spec_set=Account)
        accts = [
            Mock(
                spec_set=PlaidAccount, item_id='Item1', account_id=1,
                account=acctA
            ),
            Mock(
                spec_set=PlaidAccount, item_id='Item1', account_id=3,
                account=acctB
            ),
        ]
        mock_item = Mock(
            item_id='Item1', access_token='Token1',
            last_updated=datetime(2019, 1, 1, 1, 1, 1)
        )
        self.mock_client.item_get.return_value = {
            'item': {},
            'status': {'transactions': {'foo': 'bar'}}
        }
        txns = [
            [
                {'account_id': 1, 'amount': 10},
                {'account_id': 1, 'amount': 11},
                {'account_id': 3, 'amount': 30}
            ],
            {
                1: {'account_id': 1, 'foo': 'bar'},
                3: {'account_id': 3, 'baz': 'blam'}
            }
        ]

        mock_igr = Mock()

        def se_sfa(_, acct, *args):
            if acct['account_id'] == 1:
                return 'sid1', 1, 4
            return 'sid2', 2, 3

        with patch.multiple(
            pbm,
            db_session=DEFAULT,
            PlaidAccount=DEFAULT,
            dtnow=DEFAULT,
            ItemGetRequest=DEFAULT,
        ) as mocks:
            mocks['db_session'].query.return_value.filter. \
                return_value.all.return_value = accts
            mocks['ItemGetRequest'].return_value = mock_igr
            with patch(f'{pb}._stmt_for_acct') as m_sfa:
                m_sfa.side_effect = se_sfa
                mocks['dtnow'].return_value = datetime(2020, 5, 25, 0, 0, 0)
                with patch(f'{pb}._get_transactions') as m_gt:
                    m_gt.return_value = txns
                    res = self.cls._do_item(mock_item, 15)
        assert isinstance(res, PlaidUpdateResult)
        assert res.item == mock_item
        assert res.updated == 7
        assert res.added == 3
        assert res.success is True
        assert res.exc is None
        assert res.stmt_ids == ['sid1', 'sid2']
        assert m_sfa.mock_calls == [
            call(
                acctA,
                {'account_id': 1, 'foo': 'bar'},
                [
                    {'account_id': 1, 'amount': 10},
                    {'account_id': 1, 'amount': 11},
                ],
                datetime(2020, 5, 25, 0, 0, 0)
            ),
            call(
                acctB,
                {'account_id': 3, 'baz': 'blam'},
                [
                    {'account_id': 3, 'amount': 30}
                ],
                datetime(2020, 5, 25, 0, 0, 0)
            )
        ]
        assert mock_item.last_updated == datetime(2020, 5, 25, 0, 0, 0)
        assert mocks['db_session'].mock_calls == [
            call.query(mocks['PlaidAccount']),
            call.query().filter(False),
            call.query().filter().all(),
            call.add(mock_item),
            call.commit()
        ]
        assert mocks['ItemGetRequest'].mock_calls == [
            call(access_token='Token1')
        ]
        assert m_gt.mock_calls == [
            call(
                'Token1',
                datetime(2020, 5, 10, 0, 0, 0),
                datetime(2020, 5, 25, 0, 0, 0)
            )
        ]

    def test_acct_none(self):
        acctA = Mock(spec_set=Account)
        accts = [
            Mock(
                spec_set=PlaidAccount, item_id='Item1', account_id=1,
                account=acctA
            ),
            Mock(
                spec_set=PlaidAccount, item_id='Item1', account_id=3,
                account=None
            ),
        ]
        mock_item = Mock(
            item_id='Item1', access_token='Token1',
            last_updated=datetime(2019, 1, 1, 1, 1, 1)
        )
        self.mock_client.item_get.return_value = {
            'item': {},
            'status': {'transactions': {'foo': 'bar'}}
        }
        txns = [
            [
                {'account_id': 1, 'amount': 10},
                {'account_id': 1, 'amount': 11},
                {'account_id': 3, 'amount': 30}
            ],
            {
                1: {'account_id': 1, 'foo': 'bar'},
                3: {'account_id': 3, 'baz': 'blam'}
            }
        ]

        mock_igr = Mock()

        def se_sfa(_, acct, *args):
            if acct['account_id'] == 1:
                return 'sid1', 1, 4
            return 'sid2', 2, 3

        with patch.multiple(
            pbm,
            db_session=DEFAULT,
            PlaidAccount=DEFAULT,
            dtnow=DEFAULT,
            ItemGetRequest=DEFAULT,
        ) as mocks:
            mocks['db_session'].query.return_value.filter.\
                return_value.all.return_value = accts
            mocks['ItemGetRequest'].return_value = mock_igr
            with patch(f'{pb}._stmt_for_acct') as m_sfa:
                m_sfa.side_effect = se_sfa
                mocks['dtnow'].return_value = datetime(2020, 5, 25, 0, 0, 0)
                with patch(f'{pb}._get_transactions') as m_gt:
                    m_gt.return_value = txns
                    res = self.cls._do_item(mock_item, 15)
        assert isinstance(res, PlaidUpdateResult)
        assert res.item == mock_item
        assert res.updated == 4
        assert res.added == 1
        assert res.success is True
        assert res.exc is None
        assert res.stmt_ids == ['sid1']
        assert m_sfa.mock_calls == [
            call(
                acctA,
                {'account_id': 1, 'foo': 'bar'},
                [
                    {'account_id': 1, 'amount': 10},
                    {'account_id': 1, 'amount': 11},
                ],
                datetime(2020, 5, 25, 0, 0, 0)
            )
        ]
        assert mock_item.last_updated == datetime(2020, 5, 25, 0, 0, 0)
        assert mocks['db_session'].mock_calls == [
            call.query(mocks['PlaidAccount']),
            call.query().filter(False),
            call.query().filter().all(),
            call.add(mock_item),
            call.commit()
        ]
        assert mocks['ItemGetRequest'].mock_calls == [
            call(access_token='Token1')
        ]
        assert m_gt.mock_calls == [
            call(
                'Token1',
                datetime(2020, 5, 10, 0, 0, 0),
                datetime(2020, 5, 25, 0, 0, 0)
            )
        ]

    def test_exception(self):
        acctA = Mock(spec_set=Account)
        acctB = Mock(spec_set=Account)
        accts = [
            Mock(
                spec_set=PlaidAccount, item_id='Item1', account_id=1,
                account=acctA
            ),
            Mock(
                spec_set=PlaidAccount, item_id='Item1', account_id=3,
                account=acctB
            ),
        ]
        mock_item = Mock(
            item_id='Item1', access_token='Token1',
            last_updated=datetime(2019, 1, 1, 1, 1, 1)
        )
        self.mock_client.item_get.return_value = {
            'item': {},
            'status': {'transactions': {'foo': 'bar'}}
        }
        txns = [
            [
                {'account_id': 1, 'amount': 10},
                {'account_id': 1, 'amount': 11},
                {'account_id': 3, 'amount': 30}
            ],
            {
                1: {'account_id': 1, 'foo': 'bar'},
                3: {'account_id': 3, 'baz': 'blam'}
            }
        ]

        mock_igr = Mock()

        def se_sfa(_, acct, *args):
            if acct['account_id'] == 1:
                return 'sid1', 1, 4
            return 'sid2', 2, 3

        ex = RuntimeError('foo')
        with patch.multiple(
            pbm,
            db_session=DEFAULT,
            PlaidAccount=DEFAULT,
            dtnow=DEFAULT,
            ItemGetRequest=DEFAULT,
        ) as mocks:
            mocks['db_session'].query.return_value.filter. \
                return_value.all.return_value = accts
            mocks['ItemGetRequest'].return_value = mock_igr
            with patch(f'{pb}._stmt_for_acct') as m_sfa:
                m_sfa.side_effect = se_sfa
                mocks['dtnow'].return_value = datetime(2020, 5, 25, 0, 0, 0)
                with patch(f'{pb}._get_transactions') as m_gt:
                    m_gt.side_effect = ex
                    res = self.cls._do_item(mock_item, 15)
        assert isinstance(res, PlaidUpdateResult)
        assert res.item == mock_item
        assert res.updated == 0
        assert res.added == 0
        assert res.success is False
        assert res.exc == ex
        assert res.stmt_ids is None
        assert m_sfa.mock_calls == []
        assert mocks['db_session'].mock_calls == []
        assert mocks['ItemGetRequest'].mock_calls == [
            call(access_token='Token1')
        ]
        assert m_gt.mock_calls == [
            call(
                'Token1',
                datetime(2020, 5, 10, 0, 0, 0),
                datetime(2020, 5, 25, 0, 0, 0)
            )
        ]


class TestGetTransactions(PlaidUpdaterTester):

    def test_happy_path_no_paginate(self):
        self.mock_client.transactions_get.return_value = {
            'transactions': [
                {'trans': 1}, {'trans': 2}, {'trans': 3}
            ],
            'accounts': [
                {'account_id': 1, 'foo': 'bar'},
                {'account_id': 3, 'baz': 'blam'}
            ],
            'total_transactions': 3
        }

        mock_tgr1 = Mock()
        mock_tgro1 = Mock()

        with patch.multiple(
            pbm,
            TransactionsGetRequest=DEFAULT,
            TransactionsGetRequestOptions=DEFAULT,
        ) as mocks:
            mocks['TransactionsGetRequest'].side_effect = [mock_tgr1]
            mocks['TransactionsGetRequestOptions'].side_effect = [mock_tgro1]
            res = self.cls._get_transactions(
                'aToken', datetime(2020, 5, 10, 0, 0, 0),
                datetime(2020, 5, 25, 0, 0, 0)
            )
        assert res == (
            [{'trans': 1}, {'trans': 2}, {'trans': 3}],
            {
                1: {'account_id': 1, 'foo': 'bar'},
                3: {'account_id': 3, 'baz': 'blam'}
            }
        )
        assert mocks['TransactionsGetRequest'].mock_calls == [
            call(
                access_token='aToken',
                start_date=date(2020, 5, 10),
                end_date=date(2020, 5, 25),
                options=mock_tgro1
            )
        ]
        assert mocks['TransactionsGetRequestOptions'].mock_calls == [call()]

    def test_happy_path_paginate(self):
        self.mock_client.transactions_get.side_effect = [
            {
                'transactions': [
                    {'trans': 1}, {'trans': 2}
                ],
                'accounts': [
                    {'account_id': 1, 'foo': 'bar'},
                    {'account_id': 3, 'baz': 'blam'}
                ],
                'total_transactions': 5
            },
            {
                'transactions': [
                    {'trans': 3}, {'trans': 4}
                ],
                'accounts': [
                    {'account_id': 1, 'foo': 'bar'},
                    {'account_id': 3, 'baz': 'blam'}
                ],
                'total_transactions': 5
            },
            {
                'transactions': [
                    {'trans': 5}
                ],
                'accounts': [
                    {'account_id': 1, 'foo': 'bar'},
                    {'account_id': 3, 'baz': 'blam'}
                ],
                'total_transactions': 5
            },
        ]

        mock_tgr1 = Mock()
        mock_tgro1 = Mock()
        mock_tgr2 = Mock()
        mock_tgro2 = Mock()
        mock_tgr3 = Mock()
        mock_tgro3 = Mock()

        with patch.multiple(
            pbm,
            TransactionsGetRequest=DEFAULT,
            TransactionsGetRequestOptions=DEFAULT,
        ) as mocks:
            mocks['TransactionsGetRequest'].side_effect = [
                mock_tgr1, mock_tgr2, mock_tgr3
            ]
            mocks['TransactionsGetRequestOptions'].side_effect = [
                mock_tgro1, mock_tgro2, mock_tgro3
            ]
            res = self.cls._get_transactions(
                'aToken', datetime(2020, 5, 10, 0, 0, 0),
                datetime(2020, 5, 25, 0, 0, 0)
            )
        assert res == (
            [
                {'trans': 1}, {'trans': 2}, {'trans': 3}, {'trans': 4},
                {'trans': 5}
            ],
            {
                1: {'account_id': 1, 'foo': 'bar'},
                3: {'account_id': 3, 'baz': 'blam'}
            }
        )
        assert mocks['TransactionsGetRequest'].mock_calls == [
            call(
                access_token='aToken',
                start_date=date(2020, 5, 10),
                end_date=date(2020, 5, 25),
                options=mock_tgro1
            ),
            call(
                access_token='aToken',
                start_date=date(2020, 5, 10),
                end_date=date(2020, 5, 25),
                options=mock_tgro2
            ),
            call(
                access_token='aToken',
                start_date=date(2020, 5, 10),
                end_date=date(2020, 5, 25),
                options=mock_tgro3
            )
        ]
        assert mocks['TransactionsGetRequestOptions'].mock_calls == [
            call(), call(offset=2), call(offset=4)
        ]


class TestStmtForAcct(PlaidUpdaterTester):

    def test_credit(self):
        mock_item = Mock(institution_id='abc123')
        mock_plaid_account = Mock(
            account_type='credit', plaid_item=mock_item
        )
        mock_acct = Mock(id=5, plaid_account=mock_plaid_account)
        type(mock_acct).name = 'acct5'
        end_dt = datetime(2020, 5, 25, 0, 0, 0, tzinfo=UTC)
        txns = {
            'item': {
                'institution_id': 'abc123'
            },
            'accounts': [
                {
                    'type': 'credit',
                    'balances': {
                        'iso_currency_code': 'USD'
                    },
                    'mask': '012345'
                }
            ]
        }
        pai = {
            'mask': 'XXXX',
            'balances': {
                'iso_currency_code': 'USD'
            }
        }
        mock_stmt = Mock(id=123)
        with patch.multiple(
            pb,
            _update_bank_or_credit=DEFAULT,
            _update_investment=DEFAULT,
            _new_updated_counts=DEFAULT
        ) as mocks:
            mocks['_new_updated_counts'].return_value = (1, 2)
            with patch(f'{pbm}.OFXStatement') as m_ofxstmt:
                with patch(f'{pbm}.db_session') as m_dbsess:
                    m_ofxstmt.return_value = mock_stmt
                    res = self.cls._stmt_for_acct(mock_acct, pai, txns, end_dt)
        assert res == (123, 1, 2)
        assert mocks['_update_bank_or_credit'].mock_calls == [
            call(end_dt, mock_acct, pai, txns, mock_stmt)
        ]
        assert mocks['_update_investment'].mock_calls == []
        assert mocks['_new_updated_counts'].mock_calls == [call()]
        assert m_ofxstmt.mock_calls == [
            call(
                account_id=5,
                filename='Plaid_acct5_1590364800.ofx',
                file_mtime=end_dt,
                as_of=end_dt,
                currency='USD',
                acctid='XXXX'
            )
        ]
        assert mock_stmt.bankid == 'abc123'
        assert mock_stmt.type == 'CreditCard'
        assert m_dbsess.mock_calls == [call.commit()]

    def test_depository(self):
        mock_item = Mock(institution_id='abc123')
        mock_plaid_account = Mock(
            account_type='depository', plaid_item=mock_item
        )
        mock_acct = Mock(id=5, plaid_account=mock_plaid_account)
        type(mock_acct).name = 'acct5'
        end_dt = datetime(2020, 5, 25, 0, 0, 0, tzinfo=UTC)
        txns = {
            'item': {
                'institution_id': 'abc123'
            },
            'accounts': [
                {
                    'type': 'depository',
                    'subtype': 'checking',
                    'balances': {
                        'iso_currency_code': 'USD'
                    },
                    'mask': '012345'
                }
            ]
        }
        pai = {
            'mask': 'XXXX',
            'balances': {
                'iso_currency_code': 'USD'
            }
        }
        mock_stmt = Mock(id=123)
        with patch.multiple(
            pb,
            _update_bank_or_credit=DEFAULT,
            _update_investment=DEFAULT,
            _new_updated_counts=DEFAULT
        ) as mocks:
            mocks['_new_updated_counts'].return_value = (1, 2)
            with patch(f'{pbm}.OFXStatement') as m_ofxstmt:
                with patch(f'{pbm}.db_session') as m_dbsess:
                    m_ofxstmt.return_value = mock_stmt
                    res = self.cls._stmt_for_acct(mock_acct, pai, txns, end_dt)
        assert res == (123, 1, 2)
        assert mocks['_update_bank_or_credit'].mock_calls == [
            call(end_dt, mock_acct, pai, txns, mock_stmt)
        ]
        assert mocks['_update_investment'].mock_calls == []
        assert mocks['_new_updated_counts'].mock_calls == [call()]
        assert m_ofxstmt.mock_calls == [
            call(
                account_id=5,
                filename='Plaid_acct5_1590364800.ofx',
                file_mtime=end_dt,
                as_of=end_dt,
                currency='USD',
                acctid='XXXX'
            )
        ]
        assert mock_stmt.bankid == 'abc123'
        assert mock_stmt.type == 'Bank'
        assert m_dbsess.mock_calls == [call.commit()]

    def test_investment(self):
        mock_item = Mock(institution_id='abc123')
        mock_plaid_account = Mock(
            account_type='investment', plaid_item=mock_item
        )
        mock_acct = Mock(id=5, plaid_account=mock_plaid_account)
        type(mock_acct).name = 'acct5'
        end_dt = datetime(2020, 5, 25, 0, 0, 0, tzinfo=UTC)
        txns = {
            'item': {
                'institution_id': None
            },
            'accounts': [
                {
                    'type': 'investment',
                    'balances': {
                        'iso_currency_code': 'USD'
                    },
                    'mask': '012345'
                }
            ]
        }
        pai = {
            'mask': 'XXXX',
            'balances': {
                'iso_currency_code': 'USD'
            }
        }
        mock_stmt = Mock(id=123)
        with patch.multiple(
            pb,
            _update_bank_or_credit=DEFAULT,
            _update_investment=DEFAULT,
            _new_updated_counts=DEFAULT
        ) as mocks:
            mocks['_new_updated_counts'].return_value = (1, 2)
            with patch(f'{pbm}.OFXStatement') as m_ofxstmt:
                with patch(f'{pbm}.db_session') as m_dbsess:
                    m_ofxstmt.return_value = mock_stmt
                    res = self.cls._stmt_for_acct(mock_acct, pai, txns, end_dt)
        assert res == (123, 1, 2)
        assert mocks['_update_bank_or_credit'].mock_calls == []
        assert mocks['_update_investment'].mock_calls == [
            call(end_dt, mock_acct, pai, mock_stmt)
        ]
        assert mocks['_new_updated_counts'].mock_calls == [call()]
        assert m_ofxstmt.mock_calls == [
            call(
                account_id=5,
                filename='Plaid_acct5_1590364800.ofx',
                file_mtime=end_dt,
                as_of=end_dt,
                currency='USD',
                acctid='XXXX'
            )
        ]
        assert mock_stmt.type == 'Investment'
        assert m_dbsess.mock_calls == [call.commit()]

    def test_loan(self):
        mock_item = Mock(institution_id='abc123')
        mock_plaid_account = Mock(
            account_type='loan', plaid_item=mock_item
        )
        mock_acct = Mock(id=5, plaid_account=mock_plaid_account)
        type(mock_acct).name = 'acct5'
        end_dt = datetime(2020, 5, 25, 0, 0, 0, tzinfo=UTC)
        txns = {
            'item': {
                'institution_id': None
            },
            'accounts': [
                {
                    'type': 'loan',
                    'balances': {
                        'iso_currency_code': 'USD'
                    },
                    'mask': '012345'
                }
            ]
        }
        pai = {
            'mask': 'XXXX',
            'balances': {
                'iso_currency_code': 'USD'
            }
        }
        mock_stmt = Mock(id=123)
        with patch.multiple(
            pb,
            _update_bank_or_credit=DEFAULT,
            _update_investment=DEFAULT,
            _new_updated_counts=DEFAULT
        ) as mocks:
            mocks['_new_updated_counts'].return_value = (1, 2)
            with patch(f'{pbm}.OFXStatement') as m_ofxstmt:
                with patch(f'{pbm}.db_session') as m_dbsess:
                    m_ofxstmt.return_value = mock_stmt
                    res = self.cls._stmt_for_acct(mock_acct, pai, txns, end_dt)
        assert res == (123, 1, 2)
        assert mocks['_update_bank_or_credit'].mock_calls == []
        assert mocks['_update_investment'].mock_calls == [
            call(end_dt, mock_acct, pai, mock_stmt)
        ]
        assert mocks['_new_updated_counts'].mock_calls == [call()]
        assert m_ofxstmt.mock_calls == [
            call(
                account_id=5,
                filename='Plaid_acct5_1590364800.ofx',
                file_mtime=end_dt,
                as_of=end_dt,
                currency='USD',
                acctid='XXXX'
            )
        ]
        assert mock_stmt.type == 'Investment'
        assert m_dbsess.mock_calls == [call.commit()]

    def test_unknown_type(self):
        mock_item = Mock(institution_id='abc123')
        mock_plaid_account = Mock(
            account_type='FooBarBaz', plaid_item=mock_item
        )
        mock_acct = Mock(id=5, plaid_account=mock_plaid_account)
        type(mock_acct).name = 'acct5'
        end_dt = datetime(2020, 5, 25, 0, 0, 0, tzinfo=UTC)
        txns = {
            'item': {
                'institution_id': None
            },
            'accounts': [
                {
                    'type': 'foo',
                    'balances': {
                        'iso_currency_code': 'USD'
                    },
                    'mask': '012345'
                }
            ]
        }
        pai = {
            'mask': 'XXXX',
            'balances': {
                'iso_currency_code': 'USD'
            }
        }
        mock_stmt = Mock(id=123)
        with patch.multiple(
            pb,
            _update_bank_or_credit=DEFAULT,
            _update_investment=DEFAULT,
            _new_updated_counts=DEFAULT
        ) as mocks:
            mocks['_new_updated_counts'].return_value = (1, 2)
            with patch(f'{pbm}.OFXStatement') as m_ofxstmt:
                with patch(f'{pbm}.db_session') as m_dbsess:
                    m_ofxstmt.return_value = mock_stmt
                    with pytest.raises(RuntimeError) as exc:
                        self.cls._stmt_for_acct(mock_acct, pai, txns, end_dt)
        assert str(exc.value) == 'ERROR: Unknown account type: FooBarBaz'
        assert mocks['_update_bank_or_credit'].mock_calls == []
        assert mocks['_update_investment'].mock_calls == []
        assert mocks['_new_updated_counts'].mock_calls == []
        assert m_ofxstmt.mock_calls == [
            call(
                account_id=5,
                filename='Plaid_acct5_1590364800.ofx',
                file_mtime=end_dt,
                as_of=end_dt,
                currency='USD',
                acctid='XXXX'
            )
        ]
        assert m_dbsess.mock_calls == []


class TestNewUpdatedCounts(PlaidUpdaterTester):

    def test_none(self):
        mock_sess = Mock(dirty=[Mock()], new=['foo', 1])
        with patch(f'{pbm}.db_session', mock_sess):
            res = self.cls._new_updated_counts()
        assert res == (0, 0)

    def test_some(self):
        mock_sess = Mock(
            dirty=[Mock(), OFXTransaction()],
            new=['foo', OFXTransaction(), 1, OFXTransaction()]
        )
        with patch(f'{pbm}.db_session', mock_sess):
            res = self.cls._new_updated_counts()
        assert res == (2, 1)


class TestUpdateBankOrCredit(PlaidUpdaterTester):

    def test_happy_path(self):
        mock_stmt = Mock(avail_bal=None, avail_bal_as_of=None)
        mock_acct = Mock(id=4, negate_ofx_amounts=False)
        end_dt = datetime(2020, 5, 25, 0, 0, 0)
        acct = {
            'balances': {
                'current': '1234.5678',
                'iso_currency_code': 'USD',
                'available': 854.2903
            }
        }
        txns = [
            {
                'pending': True,
            },
            {
                'pending': False,
                'amount': 123.4567,
                'date': date(2020, 2, 23),
                'payment_meta': {
                    'reference_number': None
                },
                'name': 'Some Txn',
                'transaction_id': 'TXN001'
            },
            {
                'pending': False,
                'amount': 482.86392,
                'date': date(2020, 3, 15),
                'payment_meta': {
                    'reference_number': '456def'
                },
                'name': 'Other Txn',
                'transaction_id': 'TXN002'
            }
        ]
        with patch(f'{pbm}.db_session') as mock_db:
            with patch(f'{pbm}.upsert_record') as mock_upsert:
                self.cls._update_bank_or_credit(
                    end_dt, mock_acct, acct, txns, mock_stmt
                )
        assert mock_stmt.as_of == end_dt
        assert mock_stmt.ledger_bal == Decimal('1234.57')
        assert mock_stmt.ledger_bal_as_of == end_dt
        assert mock_stmt.avail_bal == Decimal('854.29')
        assert mock_stmt.avail_bal_as_of == end_dt
        assert mock_stmt.currency == 'USD'
        assert mock_db.mock_calls == [call.add(mock_stmt)]
        assert mock_acct.mock_calls == [
            call.set_balance(
                overall_date=end_dt,
                ledger=Decimal('1234.57'),
                ledger_date=end_dt,
                avail=Decimal('854.29'),
                avail_date=end_dt
            )
        ]
        assert mock_upsert.mock_calls == [
            call(
                OFXTransaction,
                ['account_id', 'fitid'],
                amount=Decimal('123.46'),
                date_posted=datetime(2020, 2, 23, 0, 0, 0, tzinfo=UTC),
                fitid='TXN001',
                name='Some Txn',
                account_id=4,
                statement=mock_stmt
            ),
            call(
                OFXTransaction,
                ['account_id', 'fitid'],
                amount=Decimal('482.86'),
                date_posted=datetime(2020, 3, 15, 0, 0, 0, tzinfo=UTC),
                fitid='456def',
                name='Other Txn',
                account_id=4,
                statement=mock_stmt
            )
        ]

    def test_negate_amounts(self):
        mock_stmt = Mock(avail_bal=None, avail_bal_as_of=None)
        mock_acct = Mock(id=4, negate_ofx_amounts=True)
        end_dt = datetime(2020, 5, 25, 0, 0, 0)
        acct = {
            'balances': {
                'current': '1234.5678',
                'iso_currency_code': 'USD',
                'available': 854.2903
            }
        }
        txns = [
            {
                'pending': True,
            },
            {
                'pending': False,
                'amount': 123.4567,
                'date': date(2020, 2, 23),
                'payment_meta': {
                    'reference_number': None
                },
                'name': 'Some Txn',
                'transaction_id': 'TXN001'
            },
            {
                'pending': False,
                'amount': 482.86392,
                'date': date(2020, 3, 15),
                'payment_meta': {
                    'reference_number': '456def'
                },
                'name': 'Other Txn',
                'transaction_id': 'TXN002'
            }
        ]
        with patch(f'{pbm}.db_session') as mock_db:
            with patch(f'{pbm}.upsert_record') as mock_upsert:
                self.cls._update_bank_or_credit(
                    end_dt, mock_acct, acct, txns, mock_stmt
                )
        assert mock_stmt.as_of == end_dt
        assert mock_stmt.ledger_bal == Decimal('1234.57')
        assert mock_stmt.ledger_bal_as_of == end_dt
        assert mock_stmt.avail_bal == Decimal('854.29')
        assert mock_stmt.avail_bal_as_of == end_dt
        assert mock_stmt.currency == 'USD'
        assert mock_db.mock_calls == [call.add(mock_stmt)]
        assert mock_acct.mock_calls == [
            call.set_balance(
                overall_date=end_dt,
                ledger=Decimal('1234.57'),
                ledger_date=end_dt,
                avail=Decimal('854.29'),
                avail_date=end_dt
            )
        ]
        assert mock_upsert.mock_calls == [
            call(
                OFXTransaction,
                ['account_id', 'fitid'],
                amount=Decimal('-123.46'),
                date_posted=datetime(2020, 2, 23, 0, 0, 0, tzinfo=UTC),
                fitid='TXN001',
                name='Some Txn',
                account_id=4,
                statement=mock_stmt
            ),
            call(
                OFXTransaction,
                ['account_id', 'fitid'],
                amount=Decimal('-482.86'),
                date_posted=datetime(2020, 3, 15, 0, 0, 0, tzinfo=UTC),
                fitid='456def',
                name='Other Txn',
                account_id=4,
                statement=mock_stmt
            )
        ]

    def test_no_transactions(self):
        mock_stmt = Mock(avail_bal=None, avail_bal_as_of=None)
        mock_acct = Mock(id=4)
        end_dt = datetime(2020, 5, 25, 0, 0, 0)
        acct = {
            'balances': {
                'current': '1234.5678',
                'iso_currency_code': 'USD'
            }
        }
        txns = []
        with patch(f'{pbm}.db_session') as mock_db:
            with patch(f'{pbm}.upsert_record') as mock_upsert:
                self.cls._update_bank_or_credit(
                    end_dt, mock_acct, acct, txns, mock_stmt
                )
        assert mock_stmt.as_of == end_dt
        assert mock_stmt.ledger_bal == Decimal('1234.57')
        assert mock_stmt.ledger_bal_as_of == end_dt
        assert mock_stmt.currency == 'USD'
        assert mock_db.mock_calls == [call.add(mock_stmt)]
        assert mock_acct.mock_calls == [
            call.set_balance(
                overall_date=end_dt,
                ledger=Decimal('1234.57'),
                ledger_date=end_dt,
                avail=None,
                avail_date=None
            )
        ]
        assert mock_upsert.mock_calls == []


class TestUpdateInvestment(PlaidUpdaterTester):

    def test_happy_path(self):
        mock_stmt = Mock()
        mock_acct = Mock()
        end_dt = datetime(2020, 5, 25, 0, 0, 0)
        acct = {
            'balances': {
                'current': '1234.5678',
                'iso_currency_code': 'USD'
            }
        }
        with patch(f'{pbm}.db_session') as mock_db:
            self.cls._update_investment(end_dt, mock_acct, acct, mock_stmt)
        assert mock_stmt.as_of == end_dt
        assert mock_stmt.ledger_bal == Decimal('1234.57')
        assert mock_stmt.ledger_bal_as_of == end_dt
        assert mock_stmt.currency == 'USD'
        assert mock_db.mock_calls == [call.add(mock_stmt)]
        assert mock_acct.mock_calls == [
            call.set_balance(
                overall_date=end_dt,
                ledger=Decimal('1234.57'),
                ledger_date=end_dt
            )
        ]



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from decimal import Decimal

from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.models.txn_reconcile import TxnReconcile
from selenium.webdriver.support.ui import Select
from selenium.webdriver.common.by import By
from biweeklybudget.biweeklypayperiod import BiweeklyPayPeriod
from biweeklybudget.utils import dtnow


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestBudgets(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):  # noqa
        self.baseurl = base_url
        self.get(selenium, base_url + '/budgets')

    def test_heading(self, selenium):
        heading = selenium.find_element(By.CLASS_NAME, 'navbar-brand')
        assert heading.text == 'Budgets - BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element(By.ID, 'side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element(By.ID, 'notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'

    def test_initial_data(self, selenium):
        ptable = selenium.find_element(By.ID, 'table-periodic-budgets')
        ptexts = self.tbody2textlist(ptable)
        stable = selenium.find_element(By.ID, 'table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert ptexts == [
            ['yes', 'Income (7) (income)', '$2,345.67'],
            ['yes', 'Periodic1 (1)', '$100.00'],
            ['yes', 'Periodic2 (2)', '$234.00'],
            ['NO', 'Periodic3 Inactive (3)', '$10.23']
        ]
        assert stexts == [
            ['yes', 'Standing1 (4)', '$1,284.23'],
            ['yes', 'Standing2 (5)', '$9,482.29'],
            ['NO', 'Standing3 Inactive (6)', '-$92.29']
        ]
        pelems = self.tbody2elemlist(ptable)
        selems = self.tbody2elemlist(stable)
        assert pelems[2][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(2, null)">' \
                            'Periodic2 (2)</a>'
        assert pelems[0][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(7, null)">' \
                            'Income (7)</a> <em class="text-success">' \
                            '(income)</em>'
        assert selems[0][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(4, null)">' \
                            'Standing1 (4)</a>'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestBudgetModals(AcceptanceHelper):

    def test_00_budget_modal_verify_db(self, testdb):
        b = testdb.query(Budget).get(1)
        assert b is not None
        assert b.name == 'Periodic1'
        assert b.is_periodic is True
        assert b.description == 'P1desc'
        assert b.starting_balance == Decimal('100.00')
        assert b.is_active is True
        assert b.is_income is False
        assert b.omit_from_graphs is False

    def test_01_budget_modal_populate_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element(By.XPATH, '//a[text()="Periodic1 (1)"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 1'
        assert selenium.find_element(By.ID, 'budget_frm_name').get_attribute(
            'value') == 'Periodic1'
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_periodic').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_standing').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_description').get_attribute('value') == 'P1desc'
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance').get_attribute('value') == '100'
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance_group').is_displayed()
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance').get_attribute('value') == ''
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance_group').is_displayed() is False
        assert selenium.find_element(By.ID, 'budget_frm_active').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_income').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_omit_from_graphs').is_selected() is False

    def test_02_budget_modal_update_modal(self, base_url, selenium):
        # Fill in the form
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element(By.XPATH, '//a[text()="Periodic1 (1)"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        name = selenium.find_element(By.ID, 'budget_frm_name')
        name.clear()
        name.send_keys('EditedPeriodic1')
        desc = selenium.find_element(By.ID, 'budget_frm_description')
        desc.clear()
        desc.send_keys('EditedP1desc')
        sb = selenium.find_element(By.ID, 'budget_frm_starting_balance')
        assert sb.is_displayed()
        sb.clear()
        sb.send_keys('2345.67')
        selenium.find_element(By.ID, 'budget_frm_active').click()
        assert selenium.find_element(By.ID,
                                     'budget_frm_active').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_income').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_omit_from_graphs').is_selected() is False
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Budget 1 in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'table-periodic-budgets')
        # test that updated budget was removed from the page
        ptable = selenium.find_element(By.ID, 'table-periodic-budgets')
        pelems = self.tbody2elemlist(ptable)
        assert pelems[0][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(1, null)">' \
                            'EditedPeriodic1 (1)</a>'

    def test_03_budget_modal_verify_db(self, testdb):
        b = testdb.query(Budget).get(1)
        assert b is not None
        assert b.name == 'EditedPeriodic1'
        assert b.is_periodic is True
        assert b.description == 'EditedP1desc'
        assert b.starting_balance == Decimal('2345.6700')
        assert b.is_active is False
        assert b.is_income is False
        assert b.omit_from_graphs is False

    def test_10_populate_edit_periodic_2_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element(By.XPATH, '//a[text()="Periodic2 (2)"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 2'
        assert selenium.find_element(By.ID, 'budget_frm_name').get_attribute(
            'value') == 'Periodic2'
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_periodic').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_standing').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_description').get_attribute('value') == 'P2desc'
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance').get_attribute('value') == '234'
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance_group').is_displayed()
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance').get_attribute('value') == ''
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance_group').is_displayed() is False
        assert selenium.find_element(By.ID, 'budget_frm_active').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_income').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_omit_from_graphs').is_selected() is False

    def test_11_populate_edit_periodic_3_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element(By.XPATH,
                                     '//a[text()="Periodic3 Inactive (3)"]'
                                     )
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 3'
        assert selenium.find_element(By.ID, 'budget_frm_name').get_attribute(
            'value') == 'Periodic3 Inactive'
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_periodic').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_standing').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_description').get_attribute('value') == 'P3desc'
        elem = selenium.find_element(By.ID, 'budget_frm_starting_balance')
        assert elem.get_attribute('value') == '10.23'
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance_group').is_displayed()
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance').get_attribute('value') == ''
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance_group').is_displayed() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_active').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_income').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_omit_from_graphs').is_selected() is False

    def test_12_populate_edit_standing_1_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element(By.XPATH, '//a[text()="Standing1 (4)"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 4'
        assert selenium.find_element(By.ID, 'budget_frm_name').get_attribute(
            'value') == 'Standing1'
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_periodic').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_standing').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_description').get_attribute('value') == 'S1desc'
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance').get_attribute('value') == ''
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance_group').is_displayed() is False
        elem = selenium.find_element(By.ID, 'budget_frm_current_balance')
        assert elem.get_attribute('value') == '1284.23'
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance_group').is_displayed()
        assert selenium.find_element(By.ID, 'budget_frm_active').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_income').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_omit_from_graphs').is_selected() is True

    def test_13_populate_edit_standing_2_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element(By.XPATH, '//a[text()="Standing2 (5)"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 5'
        assert selenium.find_element(By.ID, 'budget_frm_name').get_attribute(
            'value') == 'Standing2'
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_periodic').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_standing').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_description').get_attribute('value') == 'S2desc'
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance').get_attribute('value') == ''
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance_group').is_displayed() is False
        elem = selenium.find_element(By.ID, 'budget_frm_current_balance')
        assert elem.get_attribute('value') == '9482.29'
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance_group').is_displayed()
        assert selenium.find_element(By.ID, 'budget_frm_active').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_income').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_omit_from_graphs').is_selected() is False

    def test_14_populate_edit_standing_3_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element(By.XPATH,
                                     '//a[text()="Standing3 Inactive (6)"]'
                                     )
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 6'
        assert selenium.find_element(By.ID, 'budget_frm_name').get_attribute(
            'value') == 'Standing3 Inactive'
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_periodic').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_standing').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_description').get_attribute('value') == 'S3desc'
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance').get_attribute('value') == ''
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance_group').is_displayed() is False
        elem = selenium.find_element(By.ID, 'budget_frm_current_balance')
        assert elem.get_attribute('value') == '-92.29'
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance_group').is_displayed()
        assert selenium.find_element(By.ID,
                                     'budget_frm_active').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_income').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_omit_from_graphs').is_selected() is False

    def test_20_income_verify_db(self, testdb):
        b = testdb.query(Budget).get(7)
        assert b is not None
        assert b.name == 'Income'
        assert b.is_periodic is True
        assert b.description == 'IncomeDesc'
        assert b.starting_balance == Decimal('2345.67')
        assert b.is_active is True
        assert b.is_income is True
        assert b.omit_from_graphs is True

    def test_21_populate_income_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element(By.XPATH, '//a[text()="Income (7)"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 7'
        assert selenium.find_element(By.ID, 'budget_frm_name').get_attribute(
            'value') == 'Income'
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_periodic').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_standing').is_selected() is False
        elem = selenium.find_element(By.ID, 'budget_frm_description')
        assert elem.get_attribute('value') == 'IncomeDesc'
        elem = selenium.find_element(By.ID, 'budget_frm_starting_balance')
        assert elem.get_attribute('value') == '2345.67'
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance_group').is_displayed()
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance').get_attribute('value') == ''
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance_group').is_displayed() is False
        assert selenium.find_element(By.ID, 'budget_frm_active').is_selected()
        assert selenium.find_element(By.ID, 'budget_frm_income').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_omit_from_graphs').is_selected() is True

    def test_22_update_income_modal(self, base_url, selenium):
        # Fill in the form
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element(By.XPATH, '//a[text()="Income (7)"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        name = selenium.find_element(By.ID, 'budget_frm_name')
        name.clear()
        name.send_keys('EditedIncome')
        desc = selenium.find_element(By.ID, 'budget_frm_description')
        desc.send_keys('edited')
        sb = selenium.find_element(By.ID, 'budget_frm_starting_balance')
        assert sb.is_displayed()
        sb.clear()
        sb.send_keys('123.45')
        selenium.find_element(By.ID, 'budget_frm_active').click()
        assert selenium.find_element(By.ID,
                                     'budget_frm_active').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_income').is_selected() is True
        assert selenium.find_element(By.ID,
                                     'budget_frm_omit_from_graphs').is_selected() is True
        selenium.find_element(By.ID, 'budget_frm_omit_from_graphs').click()
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Budget 7 in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'table-periodic-budgets')
        # test that updated budget was removed from the page
        ptable = selenium.find_element(By.ID, 'table-periodic-budgets')
        pelems = self.tbody2elemlist(ptable)
        assert pelems[0][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(7, null)">' \
                            'EditedIncome (7)</a> ' \
                            '<em class="text-success">(income)</em>'

    def test_23_verify_income_db(self, testdb):
        b = testdb.query(Budget).get(7)
        assert b is not None
        assert b.name == 'EditedIncome'
        assert b.is_periodic is True
        assert b.description == 'IncomeDescedited'
        assert b.starting_balance == Decimal('123.45')
        assert b.is_active is False
        assert b.is_income is True
        assert b.omit_from_graphs is False

    def test_31_populate_direct_url_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets/1')
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 1'
        assert selenium.find_element(By.ID, 'budget_frm_name').get_attribute(
            'value') == 'EditedPeriodic1'
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_periodic').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_type_standing').is_selected() is False
        elem = selenium.find_element(By.ID, 'budget_frm_description')
        assert elem.get_attribute('value') == 'EditedP1desc'
        elem = selenium.find_element(By.ID, 'budget_frm_starting_balance')
        assert elem.get_attribute('value') == '2345.67'
        assert selenium.find_element(By.ID,
                                     'budget_frm_starting_balance_group').is_displayed()
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance').get_attribute('value') == ''
        assert selenium.find_element(By.ID,
                                     'budget_frm_current_balance_group').is_displayed() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_active').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_income').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_omit_from_graphs').is_selected() is False

    def test_41_add_standing_modal(self, base_url, selenium):
        # Fill in the form
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element(By.ID, 'btn_add_budget')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        name = selenium.find_element(By.ID, 'budget_frm_name')
        name.clear()
        name.send_keys('NewStanding')
        standing = selenium.find_element(By.ID, 'budget_frm_type_standing')
        standing.click()
        desc = selenium.find_element(By.ID, 'budget_frm_description')
        desc.clear()
        desc.send_keys('Newly Added Standing')
        sb = selenium.find_element(By.ID, 'budget_frm_current_balance')
        assert sb.is_displayed()
        sb.clear()
        sb.send_keys('6789.12')
        assert selenium.find_element(By.ID,
                                     'budget_frm_active').is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_income').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'budget_frm_omit_from_graphs').is_selected() is False
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Budget 8 in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'table-periodic-budgets')
        # test that updated budget was removed from the page
        stable = selenium.find_element(By.ID, 'table-standing-budgets')
        selems = self.tbody2elemlist(stable)
        assert selems[0][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(8, null)">' \
                            'NewStanding (8)</a>'

    def test_41_add_standing_verify_db(self, testdb):
        b = testdb.query(Budget).get(8)
        assert b is not None
        assert b.name == 'NewStanding'
        assert b.is_periodic is False
        assert b.description == 'Newly Added Standing'
        assert b.current_balance == Decimal('6789.12')
        assert b.is_active is True
        assert b.is_income is False
        assert b.omit_from_graphs is False

    def test_51_add_income_modal(self, base_url, selenium):
        # Fill in the form
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element(By.ID, 'btn_add_budget')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        name = selenium.find_element(By.ID, 'budget_frm_name')
        name.clear()
        name.send_keys('NewIncome')
        periodic = selenium.find_element(By.ID, 'budget_frm_type_periodic')
        periodic.click()
        desc = selenium.find_element(By.ID, 'budget_frm_description')
        desc.clear()
        desc.send_keys('Newly Added Income')
        sb = selenium.find_element(By.ID, 'budget_frm_starting_balance')
        assert sb.is_displayed()
        sb.clear()
        sb.send_keys('123.45')
        assert selenium.find_element(By.ID,
                                     'budget_frm_active').is_selected()
        income = selenium.find_element(By.ID, 'budget_frm_income')
        income.click()
        assert income.is_selected()
        assert selenium.find_element(By.ID,
                                     'budget_frm_omit_from_graphs').is_selected() is False
        selenium.find_element(By.ID, 'budget_frm_omit_from_graphs').click()
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Budget 9 in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'table-periodic-budgets')
        # test that updated budget was removed from the page
        stable = selenium.find_element(By.ID, 'table-periodic-budgets')
        selems = self.tbody2elemlist(stable)
        assert selems[2][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(9, null)">' \
                            'NewIncome (9)</a> <em class="text-success">' \
                            '(income)</em>'

    def test_52_add_income_verify_db(self, testdb):
        b = testdb.query(Budget).get(9)
        assert b is not None
        assert b.name == 'NewIncome'
        assert b.is_periodic is True
        assert b.description == 'Newly Added Income'
        assert b.starting_balance == Decimal('123.45')
        assert b.is_active is True
        assert b.is_income is True
        assert b.omit_from_graphs is True


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestBudgetTransfer(AcceptanceHelper):

    def test_1_verify_db(self, testdb):
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 4
        max_r = max([
            t.id for t in testdb.query(TxnReconcile).all()
        ])
        assert max_r == 1

    def test_2_transfer_modal(self, base_url, selenium):
        # Fill in the form
        self.get(selenium, base_url + '/budgets')
        # test that updated budget was removed from the page
        stable = selenium.find_element(By.ID, 'table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert stexts[1] == ['yes', 'Standing2 (5)', '$9,482.29']
        ptable = selenium.find_element(By.ID, 'table-periodic-budgets')
        ptexts = self.tbody2textlist(ptable)
        assert ptexts[2] == ['yes', 'Periodic2 (2)', '$234.00']
        link = selenium.find_element(By.ID, 'btn_budget_txfr')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Budget Transfer'
        assert body.find_element(By.ID,
                                 'budg_txfr_frm_date').get_attribute('value') == dtnow(
            ).strftime('%Y-%m-%d')
        amt = body.find_element(By.ID, 'budg_txfr_frm_amount')
        amt.clear()
        amt.send_keys('123.45')
        acct_sel = Select(body.find_element(By.ID, 'budg_txfr_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        from_budget_sel = Select(
            body.find_element(By.ID, 'budg_txfr_frm_from_budget')
        )
        opts = []
        for o in from_budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['7', 'Income (i)']
        ]
        assert from_budget_sel.first_selected_option.get_attribute(
            'value') == 'None'
        from_budget_sel.select_by_value('2')
        to_budget_sel = Select(
            body.find_element(By.ID, 'budg_txfr_frm_to_budget')
        )
        opts = []
        for o in from_budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['7', 'Income (i)']
        ]
        assert to_budget_sel.first_selected_option.get_attribute(
            'value') == 'None'
        to_budget_sel.select_by_value('5')
        notes = selenium.find_element(By.ID, 'budg_txfr_frm_notes')
        notes.clear()
        notes.send_keys('Budget Transfer Notes')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transactions 5 and 6' \
                                 ' in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'table-standing-budgets')
        # test that updated budget was removed from the page
        stable = selenium.find_element(By.ID, 'table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert stexts[1] == ['yes', 'Standing2 (5)', '$9,605.74']
        ptable = selenium.find_element(By.ID, 'table-periodic-budgets')
        ptexts = self.tbody2textlist(ptable)
        assert ptexts[2] == ['yes', 'Periodic2 (2)', '$234.00']

    def test_3_verify_db(self, testdb):
        desc = 'Budget Transfer - 123.45 from Periodic2 (2) to Standing2 (5)'
        t1 = testdb.query(Transaction).get(5)
        assert t1.date == dtnow().date()
        assert t1.actual_amount == Decimal('123.45')
        assert t1.budgeted_amount == Decimal('123.45')
        assert t1.description == desc
        assert t1.notes == 'Budget Transfer Notes'
        assert t1.account_id == 1
        assert t1.scheduled_trans_id is None
        assert len(t1.budget_transactions) == 1
        assert t1.budget_transactions[0].budget_id == 2
        assert t1.budget_transactions[0].amount == Decimal('123.45')
        rec1 = testdb.query(TxnReconcile).get(2)
        assert rec1.txn_id == 5
        assert rec1.ofx_fitid is None
        assert rec1.ofx_account_id is None
        assert rec1.note == desc
        t2 = testdb.query(Transaction).get(6)
        assert t2.date == dtnow().date()
        assert t2.actual_amount == Decimal('-123.45')
        assert t2.budgeted_amount == Decimal('-123.45')
        assert t2.description == desc
        assert t2.notes == 'Budget Transfer Notes'
        assert t2.account_id == 1
        assert t2.scheduled_trans_id is None
        assert len(t2.budget_transactions) == 1
        assert t2.budget_transactions[0].budget_id == 5
        assert t2.budget_transactions[0].amount == Decimal('-123.45')
        rec2 = testdb.query(TxnReconcile).get(3)
        assert rec2.txn_id == 6
        assert rec2.ofx_fitid is None
        assert rec2.ofx_account_id is None
        assert rec2.note == desc


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestBudgetTransferStoP(AcceptanceHelper):

    def test_1_verify_db(self, testdb):
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 4
        max_r = max([
            t.id for t in testdb.query(TxnReconcile).all()
        ])
        assert max_r == 1

    def test_2_transfer_modal(self, base_url, selenium, testdb):
        # Fill in the form
        self.get(selenium, base_url + '/budgets')
        # test that updated budget was removed from the page
        stable = selenium.find_element(By.ID, 'table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert stexts[1] == ['yes', 'Standing2 (5)', '$9,482.29']
        ptable = selenium.find_element(By.ID, 'table-periodic-budgets')
        ptexts = self.tbody2textlist(ptable)
        assert ptexts[2] == ['yes', 'Periodic2 (2)', '$234.00']
        pp = BiweeklyPayPeriod.period_for_date(dtnow(), testdb)
        assert pp.budget_sums[2]['allocated'] == Decimal('222.22')
        assert pp.budget_sums[2]['budget_amount'] == Decimal('234.0')
        assert pp.budget_sums[2]['remaining'] == Decimal('11.78')
        assert pp.budget_sums[2]['spent'] == Decimal('222.22')
        assert pp.budget_sums[2]['trans_total'] == Decimal('222.22')
        link = selenium.find_element(By.ID, 'btn_budget_txfr')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Budget Transfer'
        assert body.find_element(By.ID,
                                 'budg_txfr_frm_date').get_attribute('value') == dtnow(
            ).strftime('%Y-%m-%d')
        amt = body.find_element(By.ID, 'budg_txfr_frm_amount')
        amt.clear()
        amt.send_keys('123.45')
        acct_sel = Select(body.find_element(By.ID, 'budg_txfr_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        from_budget_sel = Select(
            body.find_element(By.ID, 'budg_txfr_frm_from_budget')
        )
        opts = []
        for o in from_budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['7', 'Income (i)']
        ]
        assert from_budget_sel.first_selected_option.get_attribute(
            'value') == 'None'
        from_budget_sel.select_by_value('5')
        to_budget_sel = Select(
            body.find_element(By.ID, 'budg_txfr_frm_to_budget')
        )
        opts = []
        for o in from_budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['7', 'Income (i)']
        ]
        assert to_budget_sel.first_selected_option.get_attribute(
            'value') == 'None'
        to_budget_sel.select_by_value('2')
        notes = selenium.find_element(By.ID, 'budg_txfr_frm_notes')
        notes.clear()
        notes.send_keys('Budget Transfer Notes')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transactions 5 and 6' \
                                 ' in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'table-standing-budgets')
        # test that updated budget was removed from the page
        stable = selenium.find_element(By.ID, 'table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert stexts[1] == ['yes', 'Standing2 (5)', '$9,358.84']
        ptable = selenium.find_element(By.ID, 'table-periodic-budgets')
        ptexts = self.tbody2textlist(ptable)
        assert ptexts[2] == ['yes', 'Periodic2 (2)', '$234.00']

    def test_3_verify_db(self, testdb):
        d = dtnow().date()
        pp = BiweeklyPayPeriod.period_for_date(d, testdb)
        print('Found period for %s: %s' % (d, pp))
        assert pp.budget_sums[2]['allocated'] == Decimal('98.77')
        assert pp.budget_sums[2]['budget_amount'] == Decimal('234.0')
        assert pp.budget_sums[2]['remaining'] == Decimal('135.23')
        assert pp.budget_sums[2]['spent'] == Decimal('98.77')
        assert pp.budget_sums[2]['trans_total'] == Decimal('98.77')
        desc = 'Budget Transfer - 123.45 from Standing2 (5) to Periodic2 (2)'
        t1 = testdb.query(Transaction).get(5)
        assert t1.date == dtnow().date()
        assert t1.actual_amount == Decimal('123.45')
        assert t1.budgeted_amount == Decimal('123.45')
        assert t1.description == desc
        assert t1.notes == 'Budget Transfer Notes'
        assert t1.account_id == 1
        assert t1.scheduled_trans_id is None
        assert len(t1.budget_transactions) == 1
        assert t1.budget_transactions[0].budget_id == 5
        assert t1.budget_transactions[0].amount == Decimal('123.45')
        rec1 = testdb.query(TxnReconcile).get(2)
        assert rec1.txn_id == 5
        assert rec1.ofx_fitid is None
        assert rec1.ofx_account_id is None
        assert rec1.note == desc
        t2 = testdb.query(Transaction).get(6)
        assert t2.date == dtnow().date()
        assert t2.actual_amount == Decimal('-123.45')
        assert t2.budgeted_amount == Decimal('-123.45')
        assert t2.description == desc
        assert t2.notes == 'Budget Transfer Notes'
        assert t2.account_id == 1
        assert t2.scheduled_trans_id is None
        assert len(t2.budget_transactions) == 1
        assert t2.budget_transactions[0].budget_id == 2
        assert t2.budget_transactions[0].amount == Decimal('-123.45')
        rec2 = testdb.query(TxnReconcile).get(3)
        assert rec2.txn_id == 6
        assert rec2.ofx_fitid is None
        assert rec2.ofx_account_id is None
        assert rec2.note == desc



"""add transfer relationships to transactions

Revision ID: 61e62ef50513
Revises: 32b34a664b5a
Create Date: 2017-10-26 16:13:27.227932

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy import Column, Integer, String, Date, ForeignKey
from biweeklybudget.utils import dtnow
import logging

logger = logging.getLogger(__name__)

Session = sessionmaker()

Base = declarative_base()

# revision identifiers, used by Alembic.
revision = '61e62ef50513'
down_revision = '32b34a664b5a'
branch_labels = None
depends_on = None


class Transaction(Base):
    """
    Simplified Transaction model for the data migration...
    """

    __tablename__ = 'transactions'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: date of the transaction
    date = Column(Date, default=dtnow().date())

    #: description
    description = Column(String(254), nullable=False, index=True)

    #: free-form notes
    notes = Column(String(254))

    #: ID of the account this transaction is against
    account_id = Column(Integer, ForeignKey('accounts.id'))

    #: ID of the Budget this transaction is against
    budget_id = Column(Integer, ForeignKey('budgets.id'))

    #: If the transaction is one half of a transfer, the Transaction ID of the
    #: other half/side of the transfer.
    transfer_id = Column(Integer, ForeignKey('transactions.id'))

    def __repr__(self):
        return "<Transaction(id=%s)>" % (
            self.id
        )


class Account(Base):

    __tablename__ = 'accounts'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)


class ScheduledTransaction(Base):

    __tablename__ = 'scheduled_transactions'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)


class Budget(Base):

    __tablename__ = 'budgets'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'transactions',
        sa.Column('transfer_id', sa.Integer(), nullable=True)
    )
    op.create_foreign_key(
        op.f('fk_transactions_transfer_id_transactions'),
        'transactions',
        'transactions',
        ['transfer_id'],
        ['id']
    )
    bind = op.get_bind()
    session = Session(bind=bind)
    # begin data manipulation
    last_txn = None
    for txn in session.query(Transaction).filter(
        Transaction.description.like('Budget Transfer - %')
    ).order_by(Transaction.id.asc()).all():
        if last_txn is None:
            last_txn = txn
            continue
        if (
            txn.description == last_txn.description and
            txn.date == last_txn.date and
            txn.notes == last_txn.notes and
            txn.account_id == last_txn.account_id
        ):
            # txn and last_txn are a transfer
            last_txn.transfer_id = txn.id
            txn.transfer_id = last_txn.id
            session.add(txn)
            session.add(last_txn)
            logger.warning(
                'Inferred Transfer relationship between Transactions %d and %d',
                last_txn.id, txn.id
            )
            last_txn = None
            continue
        last_txn = txn
    session.commit()
    # ### end Alembic commands ###


def downgrade():
    op.drop_constraint(
        op.f(
            'fk_transactions_transfer_id_transactions'),
        'transactions',
        type_='foreignkey'
    )
    op.drop_column('transactions', 'transfer_id')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################

This is mostly based on http://flask.pocoo.org/docs/0.12/patterns/sqlalchemy/
"""

from biweeklybudget import settings
import logging
import os
from copy import deepcopy
import warnings
import pkg_resources

from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from pymysql.err import Warning
from alembic.config import Config
from alembic.script import ScriptDirectory
from alembic.runtime.environment import EnvironmentContext
from alembic import command

from biweeklybudget.utils import in_directory

logger = logging.getLogger(__name__)

logger.debug('Creating DB engine with connection: %s',
             settings.DB_CONNSTRING)

echo = False
if os.environ.get('SQL_ECHO', '') == 'true':
    echo = True

# for Py37 warning 3135:
# "'NO_ZERO_DATE', 'NO_ZERO_IN_DATE' and
# 'ERROR_FOR_DIVISION_BY_ZERO' sql modes should be used with strict
#  mode. They will be merged with strict mode in a future release."
warnings.filterwarnings('ignore', '.*merged with strict mode.*')

# For some reason, with PyMySQL, even setting sql_mode to TRADITIONAL isn't
# raising an Exception when data is truncated. So we need to explicitly convert
# ``pymysql.err.Warning`` to an exception...
warnings.simplefilter('error', category=Warning)

engine_args = {
    'echo': echo,
    'pool_recycle': 3600
}

if settings.DB_CONNSTRING.startswith('mysql'):
    engine_args['connect_args'] = {'sql_mode': 'TRADITIONAL'}

if 'SQL_POOL_PRE_PING' in os.environ:
    engine_args['pool_pre_ping'] = True

#: The database engine object; return value of
#: :py:func:`sqlalchemy.create_engine`.
engine = create_engine(settings.DB_CONNSTRING, **engine_args)

logger.debug('Creating DB session')

#: :py:class:`sqlalchemy.orm.scoping.scoped_session` session
db_session = scoped_session(
    sessionmaker(autocommit=False, autoflush=False, bind=engine)
)

logger.debug('Setting up Base and query')

from biweeklybudget.models.base import Base  # noqa
Base.query = db_session.query_property()
from biweeklybudget.db_event_handlers import init_event_listeners  # noqa


def _alembic_get_current_rev(config, script):
    """
    Works sorta like alembic.command.current

    :param config: alembic Config
    :return: current revision
    :rtype: str
    """
    config._curr_rev = None

    def display_version(rev, _):
        for rev in script.get_all_current(rev):
            config._curr_rev = rev.cmd_format(False)
        return []

    with EnvironmentContext(
        config,
        script,
        fn=display_version
    ):
        script.run_env()
    if config._curr_rev is not None and ' ' in config._curr_rev:
        config._curr_rev = config._curr_rev.strip().split(' ')[0]
    return config._curr_rev


def init_db():
    """
    Initialize the database; call
    :py:meth:`sqlalchemy.schema.MetaData.create_all` on the metadata object.
    """
    logger.debug(
        'Initializing database with connstring: %s', settings.DB_CONNSTRING
    )
    # import all modules here that might define models so that
    # they will be registered properly on the metadata.  Otherwise
    # you will have to import them first before calling init_db()
    alembic_ini = pkg_resources.resource_filename(
        pkg_resources.Requirement.parse('biweeklybudget'),
        'biweeklybudget/alembic/alembic.ini'
    )
    topdir = os.path.abspath(
        os.path.join(os.path.dirname(alembic_ini), '..', '..')
    )
    logger.debug('Alembic configuration: %s', alembic_ini)
    with in_directory(topdir):
        alembic_config = Config(alembic_ini)
        script = ScriptDirectory.from_config(alembic_config)
        curr_rev = _alembic_get_current_rev(alembic_config, script)
        head_rev = script.get_revision("head").revision
        if curr_rev is None:
            # alembic not initialized at all; stamp with current version
            logger.warning(
                'Alembic not setup; creating all models and stamping'
            )
            logger.debug('Creating all models')
            with warnings.catch_warnings():
                # Ignore warning from MariaDB MDEV-17544 when creating tables;
                # SQLAlchemy 1.3.13 names the primary keys, but MariaDB ignores
                # the names; MariaDB >= 10.4.7 now throws a warning for this.
                warnings.filterwarnings(
                    'ignore',
                    message=r'^\(1280, "Name.*ignored for PRIMARY key\."\)$',
                    category=Warning
                )
                Base.metadata.create_all(engine)
            command.stamp(alembic_config, "head")
            logger.debug("DB stamped at %s", head_rev)
        elif curr_rev != head_rev:
            logger.warning("Alembic head is %s but this DB is at %s; "
                           "running migrations", head_rev, curr_rev)
            command.upgrade(alembic_config, "head")
            logger.info("Migrations complete")
        else:
            logger.debug('Alembic is at the correct head version (%s)',
                         curr_rev)
    logger.debug('Done initializing DB')
    init_event_listeners(db_session, engine)


def cleanup_db():
    """
    This must be called from all scripts, using

        atexit.register(cleanup_db)

    """
    logger.debug('Closing DB session')
    db_session.remove()


def upsert_record(model_class, key_fields, **kwargs):
    """
    Upsert a record in the database.

    ``key_fields`` is either a string primary key field name (a key in the
    ``kwargs`` dict) or a list or tuple of string primary key field names, for
    compound keys.

    If a record can be found matching these keys, it will be updated and
    committed. If not, a new one will be inserted. Either way, the record is
    returned.

    :py:meth:`sqlalchemy.orm.session.Session.commit` is **NOT** called.

    :param model_class: the class of model to insert/update
    :type model_class: biweeklybudget.models.base.ModelAsDict
    :param key_fields: The field name(s) (keys in ``kwargs``) that make up the
      primary key. This can be a single string, or a list or tuple of strings
      for compound keys. The values for these key fields MUST be included in
      ``kwargs``.
    :param kwargs: arguments to provide to the model class constructor, or to
      update if there is an existing record matching the key.
    :type kwargs: dict
    :return: inserted or updated record; type is an instance of ``model_class``
    """
    args = {}
    for k, v in kwargs.items():
        if isinstance(v, Base):
            args[k] = v
        else:
            args[k] = deepcopy(v)
    pkey = None
    if isinstance(key_fields, type('')):
        pkey = args[key_fields]
        del args[key_fields]
    else:
        pkey = []
        for k in key_fields:
            pkey.append(args[k])
            del args[k]
        pkey = tuple(pkey)
    logger.info('Upserting %s key=%s: %s', model_class, pkey, kwargs)
    res = db_session.query(model_class).get(pkey)
    if res is None:
        # not in DB yet; insert
        logger.debug('INSERTing (no existing record found)')
        o = model_class(**kwargs)
        db_session.add(o)
        return o
    logger.debug('Matching record exists; updating')
    for k, v in args.items():
        setattr(res, k, v)
    return res



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
from datetime import timedelta, date

#: Address to connect to Vault at, for OFX credentials
VAULT_ADDR = 'http://127.0.0.1:8200'

#: Path to read Vault token from, for OFX credentials
TOKEN_PATH = 'vault_token.txt'

#: Path to download OFX statements to, and for backfill_ofx to read them from
STATEMENTS_SAVE_PATH = os.path.expanduser('~/ofx')

#: SQLAlchemy database connection string. Note that the value given in
#: generated documentation is the value used in CI builds, not the real default.
DB_CONNSTRING = None

# MySQL connection settings
if 'DB_CONNSTRING' in os.environ:
    DB_CONNSTRING = os.environ['DB_CONNSTRING']
else:
    MYSQL_DBNAME = 'budget'
    MYSQL_HOST = 'localhost'
    MYSQL_PORT = 3306
    MYSQL_USER = 'budgetUser'
    MYSQL_PASS = 'budgetPassword'
    DB_CONNSTRING = 'mysql+pymysql://{}:{}@{}:{}/{}?charset=utf8mb4'.format(
        MYSQL_USER, MYSQL_PASS, MYSQL_HOST, MYSQL_PORT, MYSQL_DBNAME
    )

#: :py:class:`datetime.timedelta` beyond which OFX data will be considered old
STALE_DATA_TIMEDELTA = timedelta(days=2)

#: int - FOR ACCEPTANCE TESTS ONLY - This is used to "fudge" the current time
#: to the specified integer timestamp. Used for acceptance tests only. Do NOT
#: set this outside of acceptance testing.
# Tests run as 1501223084 - Friday, July 28, 2017 6:24:44 AM UTC
BIWEEKLYBUDGET_TEST_TIMESTAMP = os.environ.get('BIWEEKLYBUDGET_TEST_TIMESTAMP')
PAY_PERIOD_START_DATE = date(2017, 7, 21)

#: When listing unreconciled transactions that need to be reconciled, any
#: :py:class:`~.OFXTransaction` before this date will be ignored.
RECONCILE_BEGIN_DATE = date(2017, 1, 1)

#: Account ID to show first in dropdown lists
DEFAULT_ACCOUNT_ID = 1

#: int - Budget ID to select as default when inputting Fuel Log entries. This
#: must be the database ID of a valid budget.
FUEL_BUDGET_ID = 2

#: Plaid Client ID
PLAID_CLIENT_ID = os.environ.get('PLAID_CLIENT_ID', 'plaidCID')

#: Plaid Secret (client secret)
PLAID_SECRET = os.environ.get('PLAID_SECRET', 'plaidSecret')

#: Plaid environment name. Use 'sandbox' to test with Plaid's Sandbox
#: environment (username: user_good, password: pass_good). Use `development` to
#: test with live users and credentials and `production` to go live
PLAID_ENV = 'Sandbox'

#: PLAID_PRODUCTS is a comma-separated list of products to use when initializing
#: Link. Note that this list must contain 'assets' in order for the app to be
#: able to create and retrieve asset reports.
PLAID_PRODUCTS = 'transactions'

#: PLAID_COUNTRY_CODES is a comma-separated list of countries for which users
#: will be able to select institutions from.
PLAID_COUNTRY_CODES = 'US'

#: PLAID_USER_ID is a unique per-user ID for users of Plaid applications.
#: Since this is a single-user app, we just hard-code to "1"
PLAID_USER_ID = '1'



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
import pickle
from base64 import b64encode

import requests

from biweeklybudget.ofxapi.exceptions import DuplicateFileException

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

logger = logging.getLogger(__name__)


class OfxApiRemote(object):
    """
    Remote OFX API client, used by ofxgetter/ofxbackfiller when running on a
    remote system.
    """

    def __init__(
        self, api_base_url, ca_bundle=None, client_cert_path=None,
        client_key_path=None
    ):
        """
        Initialize a new OFX remote API client.

        :param api_base_url: base URL to the biweeklybudget installation
        :type api_base_url: str
        :param ca_bundle: Optional; local filesystem path to a SSL CA
          certificate bundle file or directory, to use for server verification.
        :type ca_bundle: str
        :param client_cert_path: Optional; local filesystem path to a SSL
          client certificate to use for authentication, if required.
        :type client_cert_path: str
        :param client_key_path: Optional; local filesystem path to key file for
          the certificate specified in ``client_cert_path``, if key is in a
          separate file. The key must be unencrypted.
        :type client_key_path: str
        """
        logger.debug(
            'New OFX API remote client; base_url=%s cert_path=%s',
            api_base_url, client_cert_path
        )
        self._base_url = api_base_url
        self._cert_path = client_cert_path
        self._key_path = client_key_path
        self._ca_bundle = ca_bundle
        self._requests_kwargs = {}
        if ca_bundle is not None:
            self._requests_kwargs['verify'] = ca_bundle
        if client_cert_path is not None:
            if client_key_path is not None:
                self._requests_kwargs['cert'] = (
                    client_cert_path, client_key_path
                )
            else:
                self._requests_kwargs['cert'] = client_cert_path

    def get_accounts(self):
        """
        Query the database for all
        :py:attr:`ofxgetter-enabled
        <biweeklybudget.models.account.Account.for_ofxgetter>`
        :py:class:`Accounts <biweeklybudget.models.account.Account>` that have
        a non-empty
        :py:attr:`biweeklybudget.models.account.Account.ofxgetter_config` and a
        non-None
        :py:attr:`biweeklybudget.models.account.Account.vault_creds_path`.
        Return a dict of string
        :py:attr:`Account name <biweeklybudget.models.account.Account.name>` to
        dict with keys:

        - ``vault_path`` - :py:attr:`~.Account.vault_creds_path`
        - ``config`` - :py:attr:`~.Account.ofxgetter_config`
        - ``id`` - :py:attr:`~.Account.id`
        - ``cat_memo`` - :py:attr:`~.Account.ofx_cat_memo_to_name`

        :return: dict of account names to configuration
        :rtype: dict
        """
        url = urljoin(self._base_url, '/api/ofx/accounts')
        logger.debug('GET ofx accounts from: %s', url)
        r = requests.get(url, **self._requests_kwargs)
        logger.debug('API Response: HTTP %d; text: %s', r.status_code, r.text)
        return r.json()

    def update_statement_ofx(self, acct_id, ofx, mtime=None, filename=None):
        """
        Update a single statement for the specified account, from an OFX file.

        :param acct_id: Account ID that statement is for
        :type acct_id: int
        :param ofx: Ofx instance for parsed file
        :type ofx: ``ofxparse.ofxparse.Ofx``
        :param mtime: OFX file modification time (or current time)
        :type mtime: datetime.datetime
        :param filename: OFX file name
        :type filename: str
        :returns: 3-tuple of the int ID of the
          :py:class:`~biweeklybudget.models.ofx_statement.OFXStatement`
          created by this run, int count of new :py:class:`~.OFXTransaction`
          created, and int count of :py:class:`~.OFXTransaction` updated
        :rtype: tuple
        :raises: :py:exc:`RuntimeError` on error parsing OFX or unknown account
          type; :py:exc:`~.DuplicateFileException` if the file (according to the
          OFX signon date/time) has already been recorded.
        """
        encodedofx = b64encode(pickle.dumps(ofx))
        encodedmtime = b64encode(pickle.dumps(mtime))
        if not isinstance(encodedofx, type('foo')):
            encodedofx = encodedofx.decode('utf-8')
            encodedmtime = encodedmtime.decode('utf-8')
        postdata = {
            'mtime': encodedmtime,
            'filename': filename,
            'acct_id': acct_id,
            'ofx': encodedofx
        }
        url = urljoin(self._base_url, '/api/ofx/statement')
        logger.debug('POST ofx statement to: %s; data: %s', url, postdata)
        r = requests.post(url, json=postdata, **self._requests_kwargs)
        logger.debug('API Response: HTTP %d; text: %s', r.status_code, r.text)
        try:
            resp = r.json()
        except Exception:
            raise RuntimeError(
                'API response could not be JSON deserialized: %s' % r.text
            )
        if r.status_code == 500:
            raise DuplicateFileException(
                resp['account_id'], resp['filename'], resp['statement_id']
            )
        if r.status_code == 400:
            raise RuntimeError('OFX API Error: %s' % resp.get('message'))
        if r.status_code != 201:
            raise RuntimeError(
                'Unknown OFX API Status Code: %d; response: %s' % (
                    r.status_code, resp
                )
            )
        # success
        logger.debug('Successfully uploaded statement: %s', resp['message'])
        return resp['statement_id'], resp['count_new'], resp['count_updated']



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from sqlalchemy import (
    Column, Integer, String, Boolean, Text, Enum, Numeric, inspect, or_,
    ForeignKeyConstraint
)
from datetime import timedelta
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship, backref
from sqlalchemy.sql.expression import null
from decimal import Decimal

from biweeklybudget.models.base import Base, ModelAsDict
from biweeklybudget.models.account_balance import AccountBalance
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.models.plaid_accounts import PlaidAccount
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.utils import dtnow
from biweeklybudget.prime_rate import PrimeRateCalculator
import json
import enum
from biweeklybudget.settings import STALE_DATA_TIMEDELTA, RECONCILE_BEGIN_DATE

logger = logging.getLogger(__name__)


class NoInterestChargedError(Exception):
    """
    Exception raised when an :py:class:`~.Account` does not have an
    OFXTransaction for interest charged within the last 32 days.
    """

    def __init__(self, acct):
        self.account = acct
        super(NoInterestChargedError, self).__init__(
            'Could not find last interest charge for account %s' % self
        )


class AcctType(enum.Enum):
    Bank = 1
    Credit = 2
    Investment = 3
    Cash = 4
    Other = 5

    @property
    def as_dict(self):
        return {'name': self.name, 'value': self.value}

    @classmethod
    def transferrable_types(self):
        """Return a list of the transferrable types."""
        return [
            self.Bank, self.Investment, self.Cash, self.Other
        ]


class Account(Base, ModelAsDict):

    __tablename__ = 'accounts'
    __table_args__ = (
        ForeignKeyConstraint(
            ['plaid_item_id', 'plaid_account_id'],
            [PlaidAccount.item_id, PlaidAccount.account_id]
        ),
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: name for the account
    name = Column(String(50), unique=True, index=True)

    #: description
    description = Column(String(254))

    #: whether or not to concatenate the OFX memo text onto the OFX name text;
    #: for banks like Chase that use the memo for run-on from the name
    ofx_cat_memo_to_name = Column(Boolean, default=False)

    #: path in Vault to read the credentials from
    vault_creds_path = Column(String(254))

    #: JSON-encoded ofxgetter configuration
    ofxgetter_config_json = Column(Text)

    #: For use in reconciling our :py:class:`~.Transaction` entries with
    #: the account's :py:class:`~.OFXTransaction` entries, whether or not to
    #: negate the OfxTransaction amount. We enter Transactions with income
    #: as negative amounts and expenses as positive amounts, but most bank
    #: OFX statements will show the opposite.
    negate_ofx_amounts = Column(Boolean, default=False)

    #: Include Transactions and OFXTransactions from this account when
    #: reconciling. Set to False to exclude accounts that are investment,
    #: payment only, or otherwise won't have a matching Transaction for each
    #: OFXTransaction.
    reconcile_trans = Column(Boolean, default=True, nullable=False)

    #: Type of account (Enum :py:class:`~.AcctType` )
    acct_type = Column(Enum(AcctType))

    #: credit limit, for credit accounts
    credit_limit = Column(Numeric(precision=10, scale=4))

    #: Finance rate (APR) for credit accounts
    apr = Column(Numeric(precision=5, scale=4))

    #: Margin added to the US Prime Rate to determine APR, for credit accounts.
    prime_rate_margin = Column(Numeric(precision=5, scale=4))

    #: Name of the :py:class:`biweeklybudget.interest._InterestCalculation`
    #: subclass used to calculate interest for this account.
    interest_class_name = Column(String(70))

    #: Name of the :py:class:`biweeklybudget.interest._MinPaymentFormula`
    #: subclass used to calculate minimum payments for this account.
    min_payment_class_name = Column(String(70))

    #: whether or not the account is active and can be used, or historical
    is_active = Column(Boolean, default=True)

    #: Relationship to all :py:class:`~.OFXStatement` for this Account
    all_statements = relationship(
        'OFXStatement', order_by='OFXStatement.as_of'
    )

    #: regex for matching transactions as interest charges
    re_interest_charge = Column(String(254))

    #: regex for matching transactions as interest paid
    re_interest_paid = Column(String(254))

    #: regex for matching transactions as payments
    re_payment = Column(String(254))

    #: regex for matching transactions as late fees
    re_late_fee = Column(String(254))

    #: regex for matching transactions as other fees
    re_other_fee = Column(String(254))

    #: Plaid Item ID for this account
    plaid_item_id = Column(String(70), nullable=True)

    #: Plaid Token for this account
    plaid_account_id = Column(String(70), nullable=True)

    #: :py:class:`~.PlaidAccount` this account is linked with
    plaid_account = relationship(
        'PlaidAccount', backref=backref('account', uselist=False),
        foreign_keys=[plaid_item_id, plaid_account_id]
    )

    def __repr__(self):
        return "<Account(id=%s, name='%s')>" % (
            self.id, self.name
        )

    @hybrid_property
    def for_ofxgetter(self):
        """
        Return whether or not this account should be handled by ofxgetter.

        :return: whether or not ofxgetter should run for this account
        :rtype: bool
        """
        return self.ofxgetter_config_json.isnot(None)

    @hybrid_property
    def plaid_configured(self):
        """
        Return whether or not this account is configured for Plaid.

        :return: whether or not this account is configured for Plaid.
        :rtype: bool
        """
        return self.plaid_account_id.isnot(None) & \
            self.plaid_item_id.isnot(None)

    @hybrid_property
    def is_budget_source(self):
        """
        Return whether or not this account should be considered a funding
        source for Budgets.

        :return: whether or not this account is a Budget funding source
        :rtype: bool
        """
        if self.acct_type == AcctType.Bank or self.acct_type == AcctType.Cash:
            return True
        return False

    @is_budget_source.expression
    def is_budget_source(cls):
        return or_(
            cls.acct_type.__eq__(AcctType.Bank),
            cls.acct_type.__eq__(AcctType.Cash),
        )

    @hybrid_property
    def is_stale(self):
        """
        Return whether or not there is stale data for this account.

        :return: whether or not data for this account is stale
        :rtype: bool
        """
        # return False if we've never seen OFX data
        if self.ofx_statement is None:
            return False
        return (dtnow() - self.ofx_statement.as_of) > STALE_DATA_TIMEDELTA

    @property
    def ofxgetter_config(self):
        """
        Return the deserialized ofxgetter_config_json dict.

        :return: ofxgetter config
        :rtype: dict
        """
        try:
            return json.loads(self.ofxgetter_config_json)
        except Exception:
            return {}

    def set_ofxgetter_config(self, config):
        """
        Set ofxgetter configuration.

        :param config: ofxgetter configuration
        :type config: dict
        """
        self.ofxgetter_config_json = json.dumps(config)

    def set_balance(self, **kwargs):
        """
        Create an AccountBalance object for this account and associate it with
        the account. Add it to the current session.
        """
        kwargs['account'] = self
        inspect(self).session.add(AccountBalance(**kwargs))

    @property
    def ofx_statement(self):
        """
        Return the latest OFXStatement for this Account.

        :return: latest OFXStatement for this Account
        :rtype: biweeklybudget.models.ofx_statement.OFXStatement
        """
        if len(self.all_statements) < 1:
            return None
        return self.all_statements[-1]

    @property
    def balance(self):
        """
        Return the latest AccountBalance object for this Account.

        :return: latest AccountBalance for this Account
        :rtype: biweeklybudget.models.account_balance.AccountBalance
        """
        sess = inspect(self).session
        res = sess.query(AccountBalance).with_parent(self).order_by(
            AccountBalance.id.desc()).limit(1).first()
        return res

    @property
    def unreconciled(self):
        """
        Return a query to match all unreconciled Transactions for this account.

        :param db: active database session to use for queries
        :type db: sqlalchemy.orm.session.Session
        :return: query to match all unreconciled Transactions
        :rtype: sqlalchemy.orm.query.Query
        """
        sess = inspect(self).session
        return sess.query(Transaction).filter(
            Transaction.reconcile.__eq__(null()),
            Transaction.date.__ge__(RECONCILE_BEGIN_DATE),
            Transaction.account_id.__eq__(self.id),
            Transaction.date.__le__(dtnow())
        )

    @property
    def unreconciled_sum(self):
        """
        Return the sum of all unreconciled transaction amounts for this account.

        :return: sum of amounts of all unreconciled transactions
        :rtype: float
        """
        total = Decimal('0.0')
        for t in self.unreconciled:
            total += t.actual_amount
        return total

    @property
    def effective_apr(self):
        """
        Return the effective APR for a credit account. If
        :py:attr:`~.prime_rate_margin` is not Null, return that added to the
        current US Prime Rate. Otherwise, return :py:attr:`~.apr`.

        :return: Effective account APR
        :rtype: decimal.Decimal
        """
        if self.prime_rate_margin is not None:
            sess = inspect(self).session
            return PrimeRateCalculator(sess).calculate_apr(
                self.prime_rate_margin
            )
        return self.apr

    @property
    def last_interest_charge(self):
        """
        Return the amount of the last interest charge for this account. Raise an
        exception if one could not be identified.

        :return: amount of last interest charge for this account
        :rtype: decimal.Decimal
        """
        sess = inspect(self).session
        for t in sess.query(OFXTransaction).filter(
            OFXTransaction.account_id.__eq__(self.id)
        ).order_by(
            OFXTransaction.date_posted.desc()
        ):
            if t.trans_type != 'debit':
                continue
            if 'interest charge' not in t.name.lower():
                continue
            if t.date_posted < (dtnow() - timedelta(days=32)):
                continue
            logger.debug(
                'Account %s found last interest charge as %s (%s)',
                self, (t.amount * -1), t
            )
            return t.amount * -1
        logger.warning('Could not find last interest charge for: %s', self)
        raise NoInterestChargedError(self)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging


def set_log_info(logger):
    """
    set logger level to INFO via :py:func:`~.set_log_level_format`.
    """
    set_log_level_format(logger, logging.INFO,
                         '%(asctime)s %(levelname)s:%(name)s:%(message)s')


def set_log_debug(logger):
    """
    set logger level to DEBUG, and debug-level output format,
    via :py:func:`~.set_log_level_format`.
    """
    set_log_level_format(
        logger,
        logging.DEBUG,
        "%(asctime)s [%(levelname)s %(filename)s:%(lineno)s - "
        "%(name)s.%(funcName)s() ] %(message)s"
    )


def set_log_level_format(logger, level, format):
    """
    Set logger level and format.

    :param logger: the logger object to set on
    :type logger: logging.Logger
    :param level: logging level; see the :py:mod:`logging` constants.
    :type level: int
    :param format: logging formatter format string
    :type format: str
    """
    formatter = logging.Formatter(fmt=format)
    logger.handlers[0].setFormatter(formatter)
    logger.setLevel(level)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
from datetime import date
from sqlalchemy.orm.session import Session
from decimal import Decimal

from biweeklybudget.biweeklypayperiod import BiweeklyPayPeriod
from biweeklybudget.models.utils import do_budget_transfer
from biweeklybudget.models.budget_model import Budget

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import Mock, patch, call
else:
    from unittest.mock import Mock, patch, call

pbm = 'biweeklybudget.models.utils'


class TestDoBudgetTransfer(object):

    def test_simple(self):
        mock_sess = Mock(spec_set=Session)
        pp = Mock(spec_set=BiweeklyPayPeriod)
        type(pp).start_date = date(2017, 1, 1)
        type(pp).end_date = date(2017, 1, 13)
        standing = Mock(spec_set=Budget, is_periodic=False)
        type(standing).id = 9
        type(standing).name = 'standingBudget'
        budg1 = Mock(spec_set=Budget)
        type(budg1).id = 1
        type(budg1).name = 'one'
        t1 = Mock()
        t2 = Mock()
        tr1 = Mock()
        tr2 = Mock()
        acct = Mock()
        desc = 'Budget Transfer - 123.45 from one (1) to standingBudget (9)'
        with patch('%s.Transaction' % pbm, autospec=True) as mock_t:
            with patch('%s.TxnReconcile' % pbm, autospec=True) as mock_tr:
                mock_t.side_effect = [t1, t2]
                mock_tr.side_effect = [tr1, tr2]
                res = do_budget_transfer(
                    mock_sess,
                    pp.start_date,
                    Decimal('123.45'),
                    acct,
                    budg1,
                    standing,
                    notes='foo'
                )
        assert res == [t1, t2]
        assert mock_t.mock_calls == [
            call(
                date=pp.start_date,
                budget_amounts={budg1: Decimal('123.45')},
                budgeted_amount=Decimal('123.45'),
                description=desc,
                account=acct,
                notes='foo',
                planned_budget=budg1
            ),
            call(
                date=pp.start_date,
                budget_amounts={standing: Decimal('-123.45')},
                budgeted_amount=Decimal('-123.45'),
                description=desc,
                account=acct,
                notes='foo',
                planned_budget=standing
            )
        ]
        assert mock_tr.mock_calls == [
            call(transaction=t1, note=desc),
            call(transaction=t2, note=desc)
        ]
        assert mock_sess.mock_calls == [
            call.add(t1),
            call.add(t2),
            call.add(t1),
            call.add(t2),
            call.add(tr1),
            call.add(tr2)
        ]



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from setuptools import setup, find_packages
from biweeklybudget.version import VERSION, PROJECT_URL

with open('README.rst') as f:
    long_description = f.read()

pyver_requires = []
dep_links = []
with open('requirements.txt') as f:
    for line in f.readlines():
        if line.startswith('#'):
            continue
        if '://' in line:
            dep_links.append(line.strip())
        else:
            pyver_requires.append(line.strip())

classifiers = [
    'Development Status :: 4 - Beta',
    'Environment :: Web Environment',
    'Framework :: Flask',
    'Intended Audience :: End Users/Desktop',
    'License :: OSI Approved :: GNU Affero General Public License '
    'v3 or later (AGPLv3+)',
    'Natural Language :: English',
    'Operating System :: POSIX',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.14',
    'Topic :: Office/Business :: Financial'
]

setup(
    name='biweeklybudget',
    version=VERSION,
    author='Jason Antman',
    author_email='jason@jasonantman.com',
    packages=find_packages(),
    package_data={
        'biweeklybudget': [
            'flaskapp/templates/*',
            'flaskapp/static/*',
            'alembic/alembic.ini',
            'alembic/env.py',
            'alembic/versions/*.py'
        ]
    },
    url=PROJECT_URL,
    description='Responsive Flask/SQLAlchemy personal finance app, '
                'specifically for biweekly budgeting.',
    long_description=long_description,
    install_requires=pyver_requires,
    dependency_links=dep_links,
    include_package_data=True,
    entry_points="""
    [console_scripts]
    loaddata = biweeklybudget.load_data:main
    ofxgetter = biweeklybudget.ofxgetter:main
    ofxbackfiller = biweeklybudget.backfill_ofx:main
    initdb = biweeklybudget.initdb:main
    wishlist2project = biweeklybudget.wishlist2project:main
    ofxclient = biweeklybudget.vendored.ofxclient.cli:run
    [flask.commands]
    rundev=biweeklybudget.flaskapp.cli_commands:rundev_command
    """,
    keywords="budget finance biweekly ofx flask responsive sqlalchemy bank",
    classifiers=classifiers
)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
import re
from distutils.spawn import find_executable
import subprocess

logger = logging.getLogger(__name__)

mysqldump_path = find_executable('mysqldump')
mysql_path = find_executable('mysql')

dump_ver_out = subprocess.check_output(
    [mysqldump_path, '--version']
).decode('utf-8')
mysqldump_is_mariadb = False
if 'mariadb' in dump_ver_out.lower():
    mysqldump_is_mariadb = True


def do_mysqldump(dumpdir, eng, with_data=True):
    """
    Shell out and mysqldump the database to the path specified by fpath.

    :param dumpdir: LocalPath specifying the directory to save dump to
    :param eng: SQLAlchemy engine to get connection information from
    :type eng: sqlalchemy.engine.Engine
    :param with_data: whether or not to include data, or just schema
    :type with_data: bool
    :raises: RuntimeError, AssertionError
    """
    assert mysqldump_path is not None
    assert eng.url.drivername == 'mysql+pymysql'
    if with_data:
        fpath = str(dumpdir.join('withdata.sql'))
        args = [
            mysqldump_path,
            '--create-options',
            '--routines',
            '--triggers',
            '--no-create-db'
        ]
    else:
        fpath = str(dumpdir.join('nodata.sql'))
        args = [
            mysqldump_path,
            '--create-options',
            '--no-data',
            '--no-create-db'
        ]
    if not mysqldump_is_mariadb:
        args.append('--column-statistics=0')
        args.append('--ssl-mode=DISABLED')
    else:
        args.append('--skip-ssl')
    args.append('--host=%s' % eng.url.host)
    args.append('--port=%s' % eng.url.port)
    args.append('--user=%s' % eng.url.username)
    if eng.url.password is not None:
        args.append('--password=%s' % eng.url.password)
    args.append(eng.url.database)
    logger.info('Running: %s', ' '.join(args))
    res = subprocess.check_output(args)
    if not with_data:
        # for no-data dumps, need to remove the AUTO_INCREMENT setters
        res = re.sub(rb"AUTO_INCREMENT=\d+\s", b'', res)
    with open(str(fpath), 'wb') as fh:
        fh.write(res)
    logger.info('Wrote %d bytes of SQL to %s', len(res), fpath)


def restore_mysqldump(dumpdir, eng, with_data=True):
    """
    Shell out and restore a mysqldump file to the database.

    :param dumpdir: LocalPath specifying the directory to save dump to
    :param eng: SQLAlchemy engine to get connection information from
    :type eng: sqlalchemy.engine.Engine
    :param with_data: whether or not to include data, or just schema
    :type with_data: bool
    :raises: RuntimeError, AssertionError
    """
    assert mysql_path is not None
    assert eng.url.drivername == 'mysql+pymysql'
    if with_data:
        fpath = str(dumpdir.join('withdata.sql'))
    else:
        fpath = str(dumpdir.join('nodata.sql'))
    args = [
        mysql_path,
        '--batch',
        '--host=%s' % eng.url.host,
        '--port=%s' % eng.url.port,
        '--user=%s' % eng.url.username,
        '--database=%s' % eng.url.database
    ]
    if not mysqldump_is_mariadb:
        args.append('--ssl-mode=DISABLED')
    else:
        args.append('--skip-ssl')
    if eng.url.password is not None:
        args.append('--password=%s' % eng.url.password)
    logger.info('Passing %s to %s', fpath, ' '.join(args))
    with open(str(fpath), 'rb') as fh:
        proc = subprocess.Popen(args, stdin=fh)
        stdout, stderr = proc.communicate()
    logger.info('MySQL dump restore complete.')
    logger.debug('mysql STDOUT: %s', stdout)
    logger.debug('mysql STDERR: %s', stderr)



"""add plaid_item_id and plaid_token attributes on Account

Revision ID: ac93b5c826a5
Revises: 073142f641b3
Create Date: 2020-01-26 20:57:22.045669

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ac93b5c826a5'
down_revision = '073142f641b3'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'accounts',
        sa.Column('plaid_item_id', sa.String(length=70), nullable=True)
    )
    op.add_column(
        'accounts',
        sa.Column('plaid_token', sa.String(length=70), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('accounts', 'plaid_token')
    op.drop_column('accounts', 'plaid_item_id')
    # ### end Alembic commands ###
